{# This Software (Dioptra) is being made available as a public service by the
National Institute of Standards and Technology (NIST), an Agency of the United
States Department of Commerce. This software was developed in part by employees of
NIST and in part by NIST contractors. Copyright in portions of this software that
were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
to Title 17 United States Code Section 105, works of NIST employees are not
subject to copyright protection in the United States. However, NIST may hold
international copyright in software created by its employees and domestic
copyright (or licensing rights) in portions of software that were assigned or
licensed to NIST. To the extent that NIST holds copyright in this software, it is
being made available under the Creative Commons Attribution 4.0 International
license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
of the software developed or licensed by NIST.

ACCESS THE FULL CC BY 4.0 LICENSE HERE:
https://creativecommons.org/licenses/by/4.0/legalcode #}

{% extends "sphinx_book_theme/layout.html" %}

{%- block extrahead %}
    {{ super() }}

    <link rel="stylesheet" href="https://pages.nist.gov/nist-header-footer/css/nist-combined.css">
    <link rel="stylesheet" type="text/css" href="https://pages.nist.gov/leaveNotice/css/jquery.leaveNotice.css" />
    <script src="https://pages.nist.gov/nist-header-footer/js/nist-header-footer.js" type="text/javascript" defer="defer"></script>
    <script async type="text/javascript" id="_fed_an_ua_tag" src="https://dap.digitalgov.gov/Universal-Federated-Analytics-Min.js?agency=NIST&subagency=github&pua=UA-66610693-1&yt=true&exts=ppsx,pps,f90,sch,rtf,wrl,txz,m1v,xlsm,msi,xsd,f,tif,eps,mpg,xml,pl,xlt,c"></script>
{%- endblock %}

{%- block footer %}
{%- include "footer.html" %}
{%- endblock %}
