"""empty message

Revision ID: a278f46a1c9d
Revises: 6a343ff9bc2f
Create Date: 2020-09-29 20:09:23.101994

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import column, table

# revision identifiers, used by Alembic.
revision = "a278f46a1c9d"
down_revision = "6a343ff9bc2f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("experiments", schema=None) as batch_op:
        batch_op.add_column(sa.Column("is_deleted", sa.Boolean(), nullable=True))

    role = table("experiments", column("is_deleted"))
    op.execute(role.update().values(is_deleted=False))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("experiments", schema=None) as batch_op:
        batch_op.drop_column("is_deleted")

    # ### end Alembic commands ###
