"""empty message

Revision ID: 018130a0bf6c
Revises: 32d5c0055e1b
Create Date: 2022-08-24 18:34:48.570620

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "018130a0bf6c"
down_revision = "32d5c0055e1b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "users",
        sa.Column(
            "user_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("username", sa.Text(), nullable=False),
        sa.Column("password", sa.Text(), nullable=False),
        sa.Column("email_address", sa.Text(), nullable=False),
        sa.Column("created_on", sa.DateTime(), nullable=False),
        sa.Column("last_modified_on", sa.DateTime(), nullable=False),
        sa.Column("last_login_on", sa.DateTime(), nullable=False),
        sa.Column("user_expire_on", sa.DateTime(), nullable=False),
        sa.Column("password_expire_on", sa.DateTime(), nullable=False),
        sa.Column("is_deleted", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("user_id", name=op.f("pk_users")),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("users")
    # ### end Alembic commands ###
