# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The schemas for serializing/deserializing the auth endpoint objects."""
from __future__ import annotations

from marshmallow import Schema, fields


class LoginSchema(Schema):
    """The request and response fields used when logging into an account."""

    username = fields.String(
        attribute="username",
        metadata=dict(description="The account username."),
        required=True,
    )
    password = fields.String(
        attribute="password",
        metadata=dict(description="The account password."),
        load_only=True,
        required=True,
    )
    status = fields.String(
        attribute="status",
        metadata=dict(description="The status of the login request."),
        dump_only=True,
    )


class LogoutSchema(Schema):
    """The response fields used when logging out of an account."""

    username = fields.String(
        attribute="username",
        metadata=dict(description="The account username."),
        dump_only=True,
        required=True,
    )
    everywhere = fields.Bool(
        attribute="everywhere",
        metadata=dict(description="Logout on all devices."),
        load_default=lambda: False,
    )
    status = fields.String(
        attribute="status",
        metadata=dict(description="The status of the logout request."),
        dump_only=True,
    )


class LogoutQueryParametersSchema(Schema):
    """The query parameters schema for logging out of an account."""

    everywhere = fields.Bool(
        attribute="everywhere",
        metadata=dict(description="Logout on all devices."),
        load_default=lambda: False,
    )
